//------------------------------------------------------------------------
//
// Author      : Maarten "McCLaw" Kronberger
// Email       : sulacomcclaw@hotmail.com
// Website     : http://www.sulaco.co.za
// Date        : 1 April 2003
// Version     : 1.0
// Description : Skeletal Character animation using Keyframe interpolation and 
//               Milkshape 3D ASCII files
//
//------------------------------------------------------------------------
unit Model;

interface
uses
  OpenGL,
  Matrix,
  Windows,
  Textures,
  SysUtils;

const
  MS_MAX_NAME = 32;
  MS_MAX_PATH = 256;

type
  clsVec = class
  public
    x,y,z : single;
    w     : single;	// 4th homogenous coordinate
    u,v   : single;	// texture position
    bone : integer;	// index into Model bones array

    { Transform All 4 coorditates }
	  procedure	transform( m : clsMatrix );

    { Transform first 3 Coordinates }
    procedure	transform3( m : clsMatrix );
  end;
type pclsVec = ^clsVec;

type
    clsTri = class
    public
      v : array [0..2] of integer;	// indices into vertices[]
	    n : array [0..2] of integer;	// indices into normals[], normal per vertex
    end;
type pclsTri = ^clsTri;

type
  clsNormal = class
  public
  	x,y,z : single;
  end;
type pclsNormal = ^clsNormal;

type
  clsShape = class
	public
    num_vertices : integer;
		vertices : array of clsVec;

		num_triangles : integer;
		triangles : array of clsTri;

		num_normals : integer;
		normals : array of clsNormal;

		{	Constructor. }
		constructor create();

    { Load From File (Not Implemented) Created For Future use }
    function loadFromFile( filename : string ) : boolean;

    { Load an Milkshape 3D Ascii File Section }
		function loadFromMs3dAsciiSegment(var ifile : TextFile) : boolean;

    { Save To File (Not Implemented) Created For Future use }
		function saveToFile( filename : string  ) : boolean;

    { Render Shape }
		procedure render();
  end;
type pclsShape = ^clsShape;

type
  clsMaterial = class
  public
    { Constructor. }
	  constructor create();

    { Load an Milkshape 3D Ascii File Section }
	  function loadFromMs3dAsciiSegment(var ifile : TextFile ) : boolean;

    { Activate Materials And Textures}
	  procedure activate();

    { Reload Textures}
	  procedure reloadTexture();

  private
	  Name : string;
  	Ambient : array [0..3] of single;
  	Diffuse : array [0..3] of single;
  	Specular : array [0..3] of single;
  	Emissive : array [0..3] of single;
  	Shininess : single;
  	Transparency : single;
  	DiffuseTexture : string;
  	AlphaTexture : string;
  	textexture : GLuint;
  end;
type pclsMaterial = ^clsMaterial;
type
  clsKeyFrame = class
	public
		Time : single;
		Value : array [0..2] of single;
  end;
type pclsKeyFrame = ^clsKeyFrame;

type
  clsBone = class
	public
    Name : string;
		ParentName : string;
		Parent : ^clsBone;						// pointer to parent bone (or NULL)
		startPosition : array [0..2] of single;
    startRotation : array [0..2] of single;
		m_relative : clsMatrix;				// fixed transformation matrix relative to parent
		m_final : clsMatrix;				  // absolute in accordance to animation
		NumPositionKeys : integer;
		PositionKeyFrames : array of clsKeyFrame;
		NumRotationKeys : integer;
		RotationKeyFrames : array of clsKeyFrame;

    { Constructor }
    constructor create();

    { Load an Milkshape 3D Ascii File Section }
		function loadFromMs3dAsciiSegment(var ifile : TextFile ) : boolean;

    { Render Bones }
		procedure render();

    { Not Implemented }
		procedure fixup();

    { Advance Animation to Current time }
		procedure advanceTo( CurrentTime : single );

    { Initialize Bone System }
		procedure initialize();
  end;
type pclsBone = ^clsBone;

type
  clsModel = class
	public

    MaxTime : single;		// length of animation (= Frames)
		CurrentTime : single;	// advancement within animation

		num_shapes : integer;
		shapes : array of clsShape;
		material_indices : array of integer;

		num_materials : integer;
		materials : array of clsMaterial;

		num_bones : integer;
		bones :  array of clsBone;

    { Constructor. }
		constructor create();

    { Load a Milkshape 3D Ascii File }
		function loadFromMs3dAsciiFile( filename : string ) : boolean;

    { Reload all Textures }
		procedure	reloadTextures();

    { Render Model}
		procedure	render();

    { Render Bones }
		procedure	renderBones();

    { Link Child And Parent Bones }
		function linkBones() : boolean;

    { Initialize All The Bones }
		procedure	initializeBones();

    { Advance Animation To DeltaTime }
		procedure	advanceAnimation( deltaTime : single );

    { Attach The Mesh To The Bones }
		procedure	attachSkin();


  end;

implementation

procedure glBindTexture(target: GLenum; texture: GLuint); stdcall; external opengl32;

{ clsVec }

{---------------------------------------------------------------------}
{  Transform All 4 coorditates                                        }
{---------------------------------------------------------------------}
procedure clsVec.transform(m: clsMatrix);
var matrix : array [0..15] of single;
    vector : array [0..3] of single;
begin

  m.getMatrix(matrix);

	vector[0] := x*matrix[0]+y*matrix[4]+z*matrix[8]+matrix[12];
	vector[1] := x*matrix[1]+y*matrix[5]+z*matrix[9]+matrix[13];
	vector[2] := x*matrix[2]+y*matrix[6]+z*matrix[10]+matrix[14];
	vector[3] := x*matrix[3]+y*matrix[7]+z*matrix[11]+matrix[15];

  x := vector[0];
	y := vector[1];
	z := vector[2];
	w := vector[3];

end;

{---------------------------------------------------------------------}
{  Transform first 3 Coordinates                                      }
{---------------------------------------------------------------------}
procedure clsVec.transform3(m: clsMatrix);
var matrix : array [0..15] of single;
    vec : array [0..2] of single;
begin
	m.getMatrix(matrix);

	vec[0] := x*matrix[0]+y*matrix[4]+z*matrix[8];
	vec[1] := x*matrix[1]+y*matrix[5]+z*matrix[9];
	vec[2] := x*matrix[2]+y*matrix[6]+z*matrix[10];

  x := vec[0];
  y := vec[1];
  z := vec[2];

end;

{ clsShape }

{---------------------------------------------------------------------}
{  Constructor.                                                       }
{---------------------------------------------------------------------}
constructor clsShape.create;
begin
  num_vertices := 0;
	vertices := Nil;
	num_triangles := 0;
	triangles := Nil;
	num_normals := 0;
	normals := Nil;
end;

{---------------------------------------------------------------------}
{  Render Shape                                                       }
{---------------------------------------------------------------------}
procedure clsShape.render;
var i,j : integer;
	  tri : clsTri;
	  vec : clsVec;
	  N   : clsNormal;
begin
//	glEnable(GL_LIGHTING);
	glBegin(GL_TRIANGLES);
    for i := 0 to num_triangles - 1	do// for each triangle
    begin
      tri := triangles[i];				// get next triangle to triangle

      for j := 0 to 2 do	// 3 vertices of the triangle
      begin
        N := normals[tri.n[j]];
        glNormal3f(N.x, N.y, N.z);	// set normal vector  (object space)

        vec := vertices[tri.v[j]];		// pointer to vertex
        glTexCoord2f (vec.u, vec.v);			// texture coordinate

        glVertex3f( vec.x, vec.y, vec.z );	// 3d coordinate (object space)
      end;
    end;
	glEnd();
end;

{---------------------------------------------------------------------}
{  Save To File (Not Implemented) Created For Future use              }
{---------------------------------------------------------------------}
function clsShape.saveToFile(filename: string): boolean;
begin
  { TODO : Look at the implementation of this for next phase }
end;

{---------------------------------------------------------------------}
{  Load From File (Not Implemented) Created For Future use            }
{---------------------------------------------------------------------}
function clsShape.loadFromFile(filename: string): boolean;
begin
  { TODO : Look at the implementation of this for next phase }
end;

{---------------------------------------------------------------------}
{  Load an Milkshape 3D Ascii File Section                            }
{---------------------------------------------------------------------}
function clsShape.loadFromMs3dAsciiSegment(var ifile: TextFile): boolean;
var nFlags, nIndex, j : integer;
begin
  //
  // vertices
  //
  Readln(ifile,num_vertices);
  if eof(ifile) then
  begin
    result := false;
    exit;
  end;
  SetLength(vertices,num_vertices);
  if (num_vertices > 0)then
  begin

    for j := 0 to num_vertices - 1 do
    begin
      //init
      vertices[j] := clsVec.Create;

      Readln(ifile,nFlags,vertices[j].x, vertices[j].y, vertices[j].z,vertices[j].u, vertices[j].v,vertices[j].bone);

      if eof(ifile) then
      begin
        result := false;
        exit;
      end;

      // adjust the y direction of the texture coordinate
      vertices[j].v := 1.0 - vertices[j].v;
    end;
  end;
  
  //
  // normals
  //
  Readln(ifile,num_normals);
  if eof(ifile) then
  begin
    result := false;
    exit;
  end;


  SetLength(normals,num_normals);
  if num_normals > 0 then
  begin
    for j := 0 to num_normals - 1 do
    begin

      //init
      normals[j] := clsNormal.Create;

      Readln(ifile,normals[j].x, normals[j].y, normals[j].z);
      if eof(ifile) then
      begin
        result := false;
        exit;
      end;

    end;
  end;

  //
  // triangles
  //
  Readln(ifile,num_triangles);

  if eof(ifile) then
  begin
    result := false;
    exit;
  end;

  SetLength (triangles,num_triangles);

  for j := 0 to num_triangles - 1 do
  begin

    triangles[j] := clsTri.Create;

    ReadLn(ifile,nFlags,triangles[j].v[0], triangles[j].v[1], triangles[j].v[2],triangles[j].n[0], triangles[j].n[1], triangles[j].n[2],nIndex);
    if eof(ifile) then
    begin
      result := false;
      exit;
    end;

    assert(triangles[j].v[0] >= 0);
    assert(triangles[j].v[0] < num_vertices);
    assert(triangles[j].v[1] >= 0);
    assert(triangles[j].v[1] < num_vertices);
    assert(triangles[j].v[2] >= 0);
    assert(triangles[j].v[2] < num_vertices);
  end;

  result := true;
end;

{ clsMaterial }

{---------------------------------------------------------------------}
{  Activate Materials And Textures                                    }
{---------------------------------------------------------------------}
procedure clsMaterial.activate;
begin
  glMaterialfv( GL_FRONT, GL_AMBIENT, @Ambient );
	glMaterialfv( GL_FRONT, GL_DIFFUSE, @Diffuse );
	glMaterialfv( GL_FRONT, GL_SPECULAR, @Specular );
	glMaterialfv( GL_FRONT, GL_EMISSION, @Emissive );
	glMaterialf( GL_FRONT, GL_SHININESS, Shininess );

	if ( textexture > 0 ) then
	begin
		glBindTexture( GL_TEXTURE_2D, textexture );
		glEnable( GL_TEXTURE_2D );
	end
	else
		glDisable( GL_TEXTURE_2D );
end;

{---------------------------------------------------------------------}
{  Constructor.                                                       }
{---------------------------------------------------------------------}
constructor clsMaterial.create;
begin
  // Empty Constructor
end;

{---------------------------------------------------------------------}
{  Load an Milkshape 3D Ascii File Section                            }
{---------------------------------------------------------------------}
function clsMaterial.loadFromMs3dAsciiSegment(var ifile: TextFile): boolean;
var szLine : string;
begin
    // name
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,szLine);
    Name := StringReplace(szLine,'"','',[rfReplaceAll]);

    // ambient
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Ambient[0], Ambient[1], Ambient[2], Ambient[3]);

    // diffuse
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Diffuse[0], Diffuse[1], Diffuse[2], Diffuse[3]);

    // specular
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Specular[0], Specular[1], Specular[2], Specular[3]);

    // emissive
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Emissive[0], Emissive[1], Emissive[2], Emissive[3]);


    // shininess
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Shininess);

    // transparency
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Transparency);

    // diffuse texture
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

	  DiffuseTexture := '';
    Readln(ifile,DiffuseTexture);

    DiffuseTexture := StringReplace(DiffuseTexture,'"','',[rfReplaceAll]);

    // alpha texture
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    AlphaTexture := '';
    Readln(ifile,AlphaTexture);

    AlphaTexture := StringReplace(AlphaTexture,'"','',[rfReplaceAll]);

	  reloadTexture();

	  result := true;
end;

{---------------------------------------------------------------------}
{  Reload Textures                                                    }
{---------------------------------------------------------------------}
procedure clsMaterial.reloadTexture;
begin
  if( length(DiffuseTexture) > 0 ) then
		LoadTexture(DiffuseTexture, textexture,false)
	else
		textexture := 0;
end;

{ clsBone }

{---------------------------------------------------------------------}
{  Constructor                                                        }
{---------------------------------------------------------------------}
constructor clsBone.create;
begin
  m_relative := clsMatrix.create;				// fixed transformation matrix relative to parent
  m_final := clsMatrix.create;				// absolute in accordance to animation
end;

{---------------------------------------------------------------------}
{  Advance Animation to Current time                                  }
{---------------------------------------------------------------------}
procedure clsBone.advanceTo(CurrentTime: single);
var i : integer;
  	deltaTime : single;
	  fraction : single;
  	Position : array [0..2] of single;
  	Rotation : array [0..2] of single;
    m_rel , m_frame : clsMatrix;
    tempm : array [0..15] of single;
begin
	//
	// Position
  //

	// Find appropriate position key frame
	i := 0;
	while ( (i < NumPositionKeys-1) and (PositionKeyFrames[i].Time < CurrentTime) ) do
		i := i + 1;

	assert(i < NumPositionKeys);


	if( i > 0 ) then
	begin
		// Interpolate between 2 key frames

		// time between the 2 key frames
		deltaTime := PositionKeyFrames[i].Time - PositionKeyFrames[i-1].Time;

		assert( deltaTime > 0 );

		// relative position of interpolation point to the keyframes [0..1]
		fraction := (CurrentTime - PositionKeyFrames[i-1].Time) / deltaTime;

		assert( fraction > 0 );
		assert( fraction < 1.0 );

		Position[0] := PositionKeyFrames[i-1].Value[0] + fraction * (PositionKeyFrames[i].Value[0] - PositionKeyFrames[i-1].Value[0]);
		Position[1] := PositionKeyFrames[i-1].Value[1] + fraction * (PositionKeyFrames[i].Value[1] - PositionKeyFrames[i-1].Value[1]);
		Position[2] := PositionKeyFrames[i-1].Value[2] + fraction * (PositionKeyFrames[i].Value[2] - PositionKeyFrames[i-1].Value[2]);
	end
	else
  begin
		Position[0] := PositionKeyFrames[i].Value[0];
    Position[1] := PositionKeyFrames[i].Value[1];
    Position[2] := PositionKeyFrames[i].Value[2];
  end;


	//
	// Rotation
	//

	// Find appropriate rotation key frame
	i := 0;
	while( (i < NumRotationKeys-1) and (RotationKeyFrames[i].Time < CurrentTime) ) do
		i := i + 1;

	assert(i < NumRotationKeys);

	if( i > 0 ) then
	begin
		// Interpolate between 2 key frames

		// time between the 2 key frames
		deltaTime := RotationKeyFrames[i].Time - RotationKeyFrames[i-1].Time;
		assert( deltaTime > 0 );

		// relative position of interpolation point to the keyframes [0..1]
		fraction := (CurrentTime - RotationKeyFrames[i-1].Time) / deltaTime;
		assert( fraction > 0 );
		assert( fraction < 1.0 );

		Rotation[0] := RotationKeyFrames[i-1].Value[0] + fraction * (RotationKeyFrames[i].Value[0] - RotationKeyFrames[i-1].Value[0]);
		Rotation[1] := RotationKeyFrames[i-1].Value[1] + fraction * (RotationKeyFrames[i].Value[1] - RotationKeyFrames[i-1].Value[1]);
		Rotation[2] := RotationKeyFrames[i-1].Value[2] + fraction * (RotationKeyFrames[i].Value[2] - RotationKeyFrames[i-1].Value[2]);
	end
	else
	begin
    Rotation[0] := RotationKeyFrames[i].Value[0];
    Rotation[1] := RotationKeyFrames[i].Value[1];
    Rotation[2] := RotationKeyFrames[i].Value[2];
	end;

	// Now we know the position and rotation for this animation frame.
	// Let's calculate the transformation matrix (m_final) for this bone...

	m_rel := clsMatrix.create;
	m_frame := clsMatrix.create;

	// Create a transformation matrix from the position and rotation of this
	// joint in the rest position
	m_rel.setRotationRadians( startRotation );
	m_rel.setTranslation( startPosition );

	// Create a transformation matrix from the position and rotation 
	// m_frame: additional transformation for this frame of the animation
	m_frame.setRotationRadians( Rotation );
	m_frame.setTranslation( Position );

	// Add the animation state to the rest position
	m_rel.postMultiply( m_frame );

	// 
	if ( Parent = nil ) then					// this is the root node
  begin
    m_rel.getMatrix(tempm);
		m_final.setMatrixValues(tempm);	// m_final := m_rel
  end
	else									// not the root node
	begin
		// m_final := parent's m_final * m_rel (matrix concatenation)
    Parent.m_final.getMatrix(tempm);
		m_final.setMatrixValues(tempm);
		m_final.postMultiply( m_rel );
	end;
end;

{---------------------------------------------------------------------}
{  Not Implemented                                                    }
{---------------------------------------------------------------------}
procedure clsBone.fixup;
begin
  {TODO : Look for the code.Is this inherent in c++?}
end;

{---------------------------------------------------------------------}
{  Initialize Bone System                                             }
{---------------------------------------------------------------------}
procedure clsBone.initialize;
var m_rel : clsMatrix;
    tempm : array [0..15] of single;
begin
  m_rel := clsMatrix.create;
	// Create a transformation matrix from the position and rotation

	m_rel.setRotationRadians( startRotation );
	m_rel.setTranslation( startPosition );

	// Each bone's final matrix is its relative matrix concatenated onto its 
	// parent's final matrix (which in turn is ....)
	//
	if ( Parent = nil ) then
  begin
    m_rel.getMatrix(tempm);
		m_final.setMatrixValues(tempm);
  end
	else
	begin
    Parent.m_final.getMatrix(tempm);
		m_final.setMatrixValues(tempm);
		m_final.postMultiply( m_rel );
	end;
end;

{---------------------------------------------------------------------}
{  Render Bones                                                       }
{---------------------------------------------------------------------}
procedure clsBone.render;
var vector , parentvector : clsVec;
begin

  vector := clsVec.Create;
  parentvector := clsVec.Create;

	vector.x := 0;
	vector.y := 0;
	vector.z := 0;
	vector.w := 1;
	vector.transform( m_final );

	if( Parent <> nil ) then
	begin
		parentvector.x := 0;
		parentvector.y := 0;
		parentvector.z := 0;
		parentvector.w := 1;
		parentvector.transform( Parent.m_final );
	end;

	glDisable( GL_TEXTURE_2D );

	// render bone as a line
	glLineWidth(1.0);
	glColor3f(1.0, 0, 0);
	glBegin(GL_LINES);
    glVertex3f( vector.x, vector.y, vector.z );
    if( Parent <> nil ) then
      glVertex3f( parentvector.x, parentvector.y, parentvector.z )
    else
      glVertex3f( vector.x, vector.y, vector.z );

	glEnd();

	// render bone-ends as fat points
	glPointSize(2.0);
	glColor3f(1.0, 0, 1.0);
	glBegin(GL_POINTS);
	glVertex3f( vector.x, vector.y, vector.z );
	if( Parent <> nil ) then
		glVertex3f( parentvector.x, parentvector.y, parentvector.z );	
	glEnd();
	glColor3f(1.0, 1.0, 1.0);
end;

{---------------------------------------------------------------------}
{  Load an Milkshape 3D Ascii File Section                            }
{---------------------------------------------------------------------}
function clsBone.loadFromMs3dAsciiSegment(var ifile: TextFile): boolean;
var szLine : string;
    j ,nFlags : integer;
begin
    // name
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,Name);

    Name := StringReplace(Name,'"','',[rfReplaceAll]) ;

    // parent name
     if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    ParentName := '';
    Readln(ifile,ParentName);

    ParentName := StringReplace(ParentName,'"','',[rfReplaceAll]) ;

    // Start Position
	  if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;
    ReadLn(ifile,nFlags,
                 startPosition[0], startPosition[1], startPosition[2],
                 startRotation[0], startRotation[1], startRotation[2]);


    // position key count
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;
    Readln(ifile,NumPositionKeys);


	  SetLength(PositionKeyFrames,NumPositionKeys);

    for j := 0 to NumPositionKeys - 1 do
    begin

      PositionKeyFrames[j] := clsKeyFrame.Create;

		  if (eof(ifile)) then
      begin
		    result := false;
        exit;
      end;

      Readln(ifile , PositionKeyFrames[j].Time,
			PositionKeyFrames[j].Value[0],
			PositionKeyFrames[j].Value[1],
			PositionKeyFrames[j].Value[2] );

    end;

    // rotation key count
    if (eof(ifile)) then
    begin
		  result := false;
      exit;
    end;

    Readln(ifile,NumRotationKeys);

	  SetLength(RotationKeyFrames,NumRotationKeys);


    for j := 0 to NumRotationKeys -1 do
    begin

		  RotationKeyFrames[j] :=  clsKeyFrame.Create;

      if (eof(ifile)) then
      begin
		    result := false;
        exit;
      end;

        Readln(ifile , RotationKeyFrames[j].Time,
        RotationKeyFrames[j].Value[0],
			  RotationKeyFrames[j].Value[1],
			  RotationKeyFrames[j].Value[2] );

    end;
	
	result := true;
end;

{ clsModel }

{---------------------------------------------------------------------}
{  Advance Animation To DeltaTime                                     }
{---------------------------------------------------------------------}
procedure clsModel.advanceAnimation(deltaTime: single);
var i : integer;
begin
	CurrentTime := CurrentTime + deltaTime;
	if( CurrentTime > MaxTime ) then	// this is a looped animation!
		CurrentTime := 1.0;		// N.B. time starts at 1, not at 0!

  for i := 0 to num_bones -1 do
	  bones[i].advanceTo( CurrentTime );

end;

{---------------------------------------------------------------------}
{  Attach The Mesh To The Bones                                       }
{---------------------------------------------------------------------}
procedure clsModel.attachSkin;
var i, j : integer;
	bone : integer;
	matrix : clsMatrix;
	v : array [0..2] of single;
begin
	// Make vertex vector relative to the bone it's attached to

	for i := 0 to num_shapes - 1 do	// for all shapes
	begin
		for j := 0 to shapes[i].num_vertices - 1 do	// for all vertices
		begin
			bone := shapes[i].vertices[j].bone;

			if( bone <> -1 ) then	// if vertex attached to a bone
			begin
				matrix := bones[bone].m_final;

				// make relative to bone position and orientation
				// modifies vertex x,y,z

				v[0] := shapes[i].vertices[j].x;
				v[1] := shapes[i].vertices[j].y;
				v[2] := shapes[i].vertices[j].z;
				matrix.inverseTranslateVect( v );
				matrix.inverseRotateVect( v );
				shapes[i].vertices[j].x := v[0];
				shapes[i].vertices[j].y := v[1];
				shapes[i].vertices[j].z := v[2];
			end;

		end;
	end;
end;

{---------------------------------------------------------------------}
{  Constructor.                                                       }
{---------------------------------------------------------------------}
constructor clsModel.create;
begin
  num_shapes := 0;
	shapes := nil;
	num_materials := 0;
	materials := nil;
end;

{---------------------------------------------------------------------}
{  Initialize All The Bones                                           }
{---------------------------------------------------------------------}
procedure clsModel.initializeBones;
var i : integer;
begin
  for i := 0 to num_bones-1 do
			bones[i].initialize;
end;

{---------------------------------------------------------------------}
{  Link Child And Parent Bones                                        }
{---------------------------------------------------------------------}
function clsModel.linkBones: boolean;
var i, j : integer;
begin
  // The relationship between child and parent bones are defined
	// by each child bone providing the name of its parent bone.
	// This function builds corresponding pointers for faster lookup.


	// Link children to parent
  for i := 0 to num_bones-1 do
	begin

		bones[i].Parent := nil;

		if( length(bones[i].ParentName) >  0 ) then	// does bone have parent?
		begin
			for j := 0 to num_bones -1 do // search for parent
			begin
				if( bones[j].Name = bones[i].ParentName) then	// j is parent of i
				begin
					bones[i].Parent := @bones[j];
					break;
				end;
			end;
			if ( bones[i].Parent = nil) then // Unable to find parent bone
      begin
        result := false;
        exit;
      end;
    end;
  end;
	result := true;
end;

{---------------------------------------------------------------------}
{  Load a Milkshape 3D Ascii File                                     }
{---------------------------------------------------------------------}
function clsModel.loadFromMs3dAsciiFile(filename: string): boolean;
var StartTime : single;
    bError : bool;
    szLine , szName , strTemp : string;
    nFrame, nFlags, nIndex, i : integer;
    fModel :  TextFile;
begin
  bError := false;

  AssignFile(fModel,filename);
  Reset(fModel);

	CurrentTime := 0;

  while ((not eof(fModel)) and (not bError)) do
  begin
    Readln(fmodel,szLine);
    if ( copy(szLine,0,2) = '//') then
      continue;

    if szLine = '' then
       continue;

    if (pos('Frames:',szLine) = 1) then
    begin
      nFrame := StrToInt(copy(szLine,8,length(szLine)));
			MaxTime := 1.0 * nFrame;		// length of animation
      continue;
    end;

    if (pos('Frame:',szLine) = 1) then
    begin
      nFrame := StrToInt(copy(szLine,7,length(szLine)));
  		StartTime := 1.0 * nFrame;
      continue;
    end;

    if (pos('Meshes: ',szLine) = 1) then
    begin
      num_shapes := StrToInt(StringReplace(szLine,'Meshes: ','',[rfReplaceAll]));
      SetLength(shapes,num_shapes);
      SetLength(material_indices, num_shapes);
      for i := 0 to num_shapes - 1 do
      begin
			  shapes[i] := clsShape.create;
        material_indices[num_shapes-1] := 0;
      end;
      
      for i := 0 to num_shapes - 1 do
      begin

          // mesh: name, flags, material index
          Readln(fmodel,szName);
          if (eof(fmodel)) then
          begin
              bError := true;
              break;
          end;

          //remove "
          szName := StringReplace(szName,'"','',[rfReplaceAll]);

          //get nFlags
          strTemp := szName;
          strTemp := copy(strTemp,pos(' ',strTemp)+1,length(strTemp));
          nFlags := StrToInt(copy(strTemp,0,pos(' ',strTemp)-1));
          //get nIndex
          nIndex := StrToInt(copy(strTemp,pos(' ',strTemp)+1,length(strTemp)));

				  material_indices[i] := nIndex;

				  if( not shapes[i].loadFromMs3dAsciiSegment(fModel) ) then
          begin
              bError := true;
              break;
          end;
			end;
      continue;
		end;

    //
    // materials
    //
    if (pos('Materials: ',szLine) = 1) then
    begin


      num_materials := StrToInt(StringReplace(szLine,'Materials: ','',[rfReplaceAll]));

      SetLength(materials,num_materials);

      // Init materials
      for i := 0 to num_materials -1 do
        materials[i] := clsMaterial.create;

      for i := 0 to num_materials -1 do
      begin

 				if( not materials[i].loadFromMs3dAsciiSegment(fModel) ) then
        begin
            bError := true;
            break;
        end;
			end;
      continue;
    end;

		//
    // bones
    //
    if (pos('Bones: ',szLine) = 1) then
    begin


        num_bones := StrToInt(StringReplace(szLine,'Bones: ','',[rfReplaceAll]));

        SetLength(bones,num_bones);

        // Initialize Bones
        for i := 0 to num_bones -1 do
          bones[i] := clsBone.create;

        for i := 0 to num_bones -1 do
        begin

            if( not bones[i].loadFromMs3dAsciiSegment(fModel) ) then
            begin
                bError := true;
                break;
            end;
        end;
        continue;
    end;
  end;

	CloseFile(fModel);

	if( not linkBones() ) then
  begin
    result := false;
    exit;
  end;
	initializeBones();
	attachSkin();
	advanceAnimation( StartTime );
	result := true;
end;

{---------------------------------------------------------------------}
{  Reload all Textures                                                }
{---------------------------------------------------------------------}
procedure clsModel.reloadTextures;
var i : integer;
begin
	for i := 0 to num_materials - 1 do	// for each shape
		materials[i].reloadTexture();

end;

{---------------------------------------------------------------------}
{  Render Model                                                       }
{---------------------------------------------------------------------}
procedure clsModel.render;
var k,i,j,materialIndex : integer;
	tri : clsTri;
  N : clsNormal;
	vec : clsVec;
  bone : clsBone;
	v : array [0..2] of single;
	matrix : clsMatrix;
begin

	for k := 0 to num_shapes - 1 do	// for each shape
	begin
		materialIndex := material_indices[k];
		if ( materialIndex >= 0 ) then
			materials[materialIndex].activate()
		else
    begin
			// Material properties?
			glDisable( GL_TEXTURE_2D );
		end;


		glBegin(GL_TRIANGLES);
		for i := 0 to shapes[k].num_triangles - 1 do	// for each triangle
		begin
			tri := shapes[k].triangles[i];				// pointer to triangle

			for j := 0 to 2 do	// 3 vertices of the triangle
			begin
        N := shapes[k].normals[tri.n[j]];
        glNormal3f(N.x, N.y, N.z);	// set normal vector  (object space)

        vec := shapes[k].vertices[tri.v[j]];		// pointer to vertex
        glTexCoord2f (vec.u, vec.v);			// texture coordinate

        if( vec.bone = -1 ) then
        begin
          glVertex3f( vec.x, vec.y, vec.z );
        end
        else
        begin
          bone := bones[vec.bone];
          matrix := bone.m_final;

          v[0] := vec.x;
          v[1] := vec.y;
          v[2] := vec.z;
          matrix.rotateVect( v );
          matrix.translateVect( v );
          glVertex3fv( @v );
        end;
			end;
		end;
		glEnd();
	end;
end;

{---------------------------------------------------------------------}
{  Render Bones                                                       }
{---------------------------------------------------------------------}
procedure clsModel.renderBones;
var i : integer;
begin
	for i := 0 to num_bones - 1 do	// for each bone
	  bones[i].render();

end;

end.

